""" TextDataset class module docstring

"""

import numpy as np
import tensorflow as tf
from . import dataset

class TextDataset(dataset.Dataset):

  @staticmethod
  @tf.function
  def normalise_text(data:tf.Tensor, label_index:int=None) -> tf.Tensor:
    """ ```data``` must have dimension rows be dataset and cols be datapoints 
    """
    min_data_vector = tf.reduce_min(data, axis=0)
    max_data_vector = tf.reduce_max(data, axis=0)

    if label_index is not None:
      min_data_vector[label_index]= 0
      max_data_vector[label_index]= 1

    numerator = tf.subtract(data, min_data_vector)
    denominator = tf.subtract(max_data_array, min_data_array)

    return tf.divide(numerator, denominator)

  @staticmethod
  @tf.function
  def normalise_labeled_text(data:tf.Tensor) -> tf.Tensor:
    return normalise_data(data, label_index=-1)

  @staticmethod
  @tf.function
  def shift_labels(data:tf.Tensor, 
                   label_index:int=-1, shift_amount:int=1
                   ) -> tf.Tensor:
    vector_shape = data.shape[1]
    if index < 0:
      index = vector_shape + index

    shifter = tf.constant([shift_amount])
    padding = tf.constant([[index, vector_shape - 1 - index]])

    subtractor = tf.pad(shifter, padding, "CONSTANT", constant_values=0)

    return tf.subtract(data, sub)




""" Main Module Doc String

"""

import os
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2" # pylint: disable=C0413

import datetime
import numpy as np
import tensorflow as tf

from ganondorf.data import datasets as gds
from ganondorf import pix2pix as p2p

def dir_path(dirname: str, *args) -> str:
  return os.path.join(os.path.dirname(__file__), dirname, *args)

def patch_dataset(pre, post):
  pre_patches  = gds.split_into_patches(pre)
  post_patches = gds.split_into_patches(post)

  return (pre_patches, post_patches)

def image_norm(image):
  max_image = np.amax(image, axis=0)
  max_image[max_image == 0] = 1
  return image / max_image

def invert(image):
  image = image * (-1)
  image = image + 1
  return image


if __name__ == "__main__":
  btl = gds.BrainTumorProgressionLoader()

  BUFFER_SIZE = 400 #1000
  BATCH_SIZE = 1
  IMG_WIDTH = 256
  IMG_HEIGHT = 256
  EPOCHS = 150  #100,000 in paper and another 100,000 after pretraining :O

  train_pre  = tf.data.Dataset.from_tensor_slices(
      list(map(image_norm, btl.pre_scans[:10]))
      )
  # train_post = tf.data.Dataset.from_tensor_slices(
    # list(map(image_norm, btl.post_scans[:10]))
    # )

  test_pre  = tf.data.Dataset.from_tensor_slices(
      list(map(image_norm, btl.pre_scans[10:]))
      )
  # test_post = tf.data.Dataset.from_tensor_slices(
    # list(map(image_norm, btl.post_scans[10:]))
    # )

  ##################################################
  # New Invert of target do make it look diff
  train_post_images = list(
      map(invert,
          map(image_norm,
              btl.post_scans[:10]
              )
         )
      )

  test_post_image = list(
      map(invert,
          map(image_norm,
              btl.post_scans[10:]
              )
         )
      )

  train_post = tf.data.Dataset.from_tensor_slices(train_post_images)

  test_post = tf.data.Dataset.from_tensor_slices(test_post_image)

  ##################################################



  train_dataset = tf.data.Dataset.zip((train_pre, train_post))
  test_dataset  = tf.data.Dataset.zip((test_pre, test_post))

  train_dataset = train_dataset.shuffle(BUFFER_SIZE)
  test_dataset  = test_dataset.shuffle(BUFFER_SIZE)

  #Ignore Batching for now due to patch splitting
  #train_dataset = train_dataset.batch(BATCH_SIZE)
  #test_dataset  = test_dataset.batch(BATCH_SIZE)


  train_dataset = train_dataset.map(patch_dataset)
  test_dataset  = test_dataset.map(patch_dataset)

  ##TODO: split and sew code!
  #>>> ds2 = ds1.map(gds.split_into_patches)
  #>>> ids2 = iter(ds2)
  #>>> x = next(ids2)
  #>>> ids2 = iter(ds2)
  #>>> x[0].shape
  #TensorShape([24, 32, 32, 1])
  #>>> xten = tf.stack(x)
  #>>> xten.shape
  #TensorShape([64, 24, 32, 32, 1])

  generator = p2p.Generator()
  discriminator = p2p.Discriminator()

  generator_optimizer = tf.keras.optimizers.Adam(learning_rate=0.0001,
                                                 beta_1=0.5)

  discriminator_optimizer = tf.keras.optimizers.Adam(learning_rate=0.0001,
                                                     beta_1=0.5)

  checkpoint_dir = dir_path("tensorflow_outputs", "training_checkpoints")

  checkpoint_prefix = os.path.join(checkpoint_dir, "ckpt")
  checkpoint = tf.train.Checkpoint(
      generator_optimizer=generator_optimizer,
      discriminator_optimizer=discriminator_optimizer,
      generator=generator,
      discriminator=discriminator)


  log_dir = dir_path("tensorflow_outputs", "logs")

  summary_writer = tf.summary.create_file_writer(
      os.path.join(log_dir,
                   "fit",
                   datetime.datetime.now().strftime("%Y%m%d-%H;%M;%S"))
      )

  p2p.fit(generator=generator,
          generator_loss=p2p.generator_loss,
          generator_optimizer=generator_optimizer,
          discriminator=discriminator,
          discriminator_loss=p2p.discriminator_loss,
          discriminator_optimizer=discriminator_optimizer,
          train_dataset=train_dataset,
          test_dataset=test_dataset,
          epochs=EPOCHS,
          checkpoint=checkpoint,
          summary_writer=summary_writer,
          checkpoint_prefix=checkpoint_prefix)




""" Module for Formating images to use with the machine learning module

"""
import PIL
import SimpleITK as sitk
import numpy as np
import os
from typing import Sequence
import tensorflow as tf
from ganondorf.core import datacore

__all__ = ['window_level', 'image_as_array', 'medical_as_array', 'as_mask',
           'split_into_patches', 'sew_patches', 'resize_nii',
           'convert_dir_images_to_nii', 'resize_medical_image', 'resize_image',
           'fix_aspect_ratio' , 'fix_med_aspect_ratio', 'resize_image_and_save',
           'fix_aspect_ratio_and_save', 'normalize', 'square_images',
           'save_image_array', 'square_pad', 'load_image_array']

def window_level(hound_image: np.array, window: int, level:int) -> np.array:

  image = hound_image.copy()  # ?????????

  if image.ndim == 3:
    half_window = window // 2
    window_min = level - half_window
    window_max = level + half_window


    image[image < window_min] = window_min
    image[image > window_max] = window_max

    image = image - window_min

    batch_count = image.shape[0]

    max_vector = \
        np.amax(np.amax(image, axis=1), axis=1) \
        .reshape(batch_count, -1, 1)

    max_matrix = np.full_like(image, max_vector)

    max_matrix[max_matrix == 0] = -1

    image = image * (255 / max_matrix)

  else:
    image = datacore.window_level(image, window, level)

  return image


#def signed_hex(num: int, bits: int = 16) -> str:
#  pass

#def color_level(hound_image: np.array) -> np.array:
#  pass

def image_as_array(image: PIL.Image.Image, mode: str=None) -> np.array:
  if mode is not None:
    image = image.convert(mode=mode)
  
  arr = np.asarray(image)
  if mode == '1':
    arr = arr.astype(np.uint8)
    arr[arr > 0] = 1

  if arr.ndim == 2:
    arr = arr[..., np.newaxis]
  return arr

def medical_as_array(image: sitk.Image) -> np.array:
  return sitk.GetArrayFromImage(image).astype(np.float32)[..., np.newaxis]

def as_mask(img):
  arr = np.asarray(img)
  arr = arr.copy()
  arr[arr > 0] = 255
  img = Image.fromarray(arr)
  img = img.convert(mode='1')
  return img

def split_into_patches(image: np.array,
                       patch_size: tuple[int, int, int] = (24,32,32)
                       ) -> list[np.array]:

  slices, height, width = image.shape[:3]
  (slice_patch, height_patch, width_patch) = patch_size

  slice_range  = math.ceil(slices / slice_patch)
  height_range = math.ceil(height / height_patch)
  width_range  = math.ceil(width  / width_patch)

  patches = []

  for i in range(slice_range):
    slice_start = i * slice_patch
    slice_end   = (i + 1) * slice_patch
    for j in range(height_range):
      height_start = j * height_patch
      height_end   = (j + 1) * height_patch
      for k in range(width_range):
        width_start = k * width_patch
        width_end   = (k + 1) * width_patch

        patches.append(image[slice_start  : slice_end,\
                             height_start : height_end,\
                             width_start  : width_end])

  return patches


def sew_patches(patches: list[np.array],
                image_size: tuple[int, int, int] = (24,256,256)
                ) -> np.array:

  height, width = image_size[1:]
  (height_patch, width_patch) = patches[0].shape[1:3]

  height_range = math.ceil(height / height_patch)
  width_range  = math.ceil(width  / width_patch)

  width_counts = len(patches) // width_range

  width_patches = \
      [np.concatenate(patches[i * width_range : width_range * (i + 1)], axis=2)\
       for i in range(width_counts)]

  height_counts = len(width_patches) // height_range

  height_patches = \
      [np.concatenate(
          width_patches[i * height_range : (i+1) * height_range],
          axis=1
          ) for i in range(height_counts)]

  return np.concatenate(height_patches, axis=0)

def resize_nii(image_filename: str,
               size: Sequence[int],
               outname: str = "out.nii",
               method=tf.image.ResizeMethod.NEAREST_NEIGHBOR) -> None:
  arr = sitk.GetArrayFromImage(sitk.ReadImage(image_filename))
  dtype = arr.dtype

  if arr.ndim == 3:
    arr = arr[..., np.newaxis]

  tensor = tf.convert_to_tensor(arr, dtype=dtype)
  image = tf.image.resize(tensor, size, method=method)

  sitk.WriteImage(sitk.GetImageFromArray(image.numpy()), outname)

def convert_dir_images_to_nii(outname: str = "out.nii",
                              dirname: str = ".") -> None:
  image_filenames = list(
      map(
          lambda fname: os.path.join(dirname, fname), os.listdir(dirname)
          )
      )

  arr = sitk.GetArrayFromImage(sitk.ReadImage(image_filenames[0]))
  for i in range(1, len(image_filenames)):
    arr = np.vstack(
        (arr, sitk.GetArrayFromImage(sitk.ReadImage(image_filenames[i])))
        )

    print(arr.shape)
  img = sitk.GetImageFromArray(arr)
  sitk.WriteImage(img, outname, imageIO="NiftiImageIO")

def resize_medical_image(image_name: str,
                 new_size: tuple[int, int] = (256, 256),
                 interpolator = sitk.sitkNearestNeighbor,
                 ) -> sitk.SimpleITK.Image:
  """ Resizes Images to a new shape preserving slice count

  """
  try:
    image = sitk.ReadImage(image_name)
  except RuntimeError as re:
    print(re)
    return None

  orig_size = image.GetSize()
  orig_spacing = image.GetSpacing()

  resize = (new_size[0], new_size[1], orig_size[2]) if len(orig_size) == 3 \
      else new_size

  new_spacing = [
      ((orig_size[0] - 1) * orig_spacing[0] / (new_size[0] - 1)),
      ((orig_size[1] - 1) * orig_spacing[1] / (new_size[1] - 1)),
      1.0
      ]

  out_image = sitk.Resample(image1=image,
                            size=resize,
                            interpolator=interpolator,
                            transform=sitk.Transform(),
                            outputOrigin=image.GetOrigin(),
                            outputDirection=image.GetDirection(),
                            outputSpacing=new_spacing)
  return out_image


def resize_image(image_name: str, #  Not always going to be str
                 new_size: tuple[int, int] = (256, 256),
                 interpolator = PIL.Image.NEAREST,
                 roi: tuple[float, float, float, float] = None,
                 reducing_gap:float=None
                 ) -> PIL.Image.Image:
  image = PIL.Image.open(image_name)
  return image.resize(new_size, interpolator, roi, reducing_gap)


def fix_aspect_ratio(img: np.array) -> np.array:
  if img.ndim == 2:
    h, w = img.shape
    img = img[np.newaxis, ...]
  elif img.ndim < 5:
    h, w = img.shape[1:3]
  else:
    h, w = img.shape[-2:img.ndim]

  (left, right, top, bottom) = (0,0,0,0)

  if h > w:
    diff = h - w
    left = diff // 2
    right = diff - left
  elif h < w:
    diff = w - h
    top = diff // 2
    bottom = diff - top
  else:
    return img

  if img.ndim == 2:
    pad = ((top, bottom), (left,right))
  else:
    extra_dims = img.ndim - 3 # 2 for width and height and 1 for leading
    pad = ((top, bottom), (left,right), *([(0,0)] * extra_dims))

  out_arrs = [np.pad(arr, pad, constant_values=0) for arr in img]

  return np.stack(out_arrs, 0)


def fix_med_aspect_ratio(img_name: str) -> sitk.SimpleITK.Image:
  img = sitk.GetArrayFromImage(sitk.ReadImage(img_name))

  return sitk.GetImageFromArray(fix_aspect_ratio(img))


def resize_image_and_save(image_name: str,
                          new_size: tuple[int, int] = (256, 256),
                          interpolator = sitk.sitkNearestNeighbor,
                          save_name: str = None) -> None:

  save_name = save_name if save_name is not None else image_name

  image = resize_image(image_name, new_size, interpolator)
  
  if iamge is not None:
    sitk.WriteImage(image, save_name)

def fix_aspect_ratio_and_save(img_name: str, save_name: str = None) -> None:

  save_name = save_name if save_name is not None else img_name
  image = fix_aspect_ratio(img_name)

  sitk.WriteImage(image, save_name)


def load_image_array(filename: str) -> np.array:
  return sitk.GetArrayFromImage(sitk.ReadImage(filename))

def save_image_array(image: np.array, filename: str = "output.nii.gz") -> None:
  sitk.WriteImage(sitk.GetImageFromArray(image), filename)

def square_pad(image: np.array)-> np.array:
  (l, r, t, b) = (0,0,0,0)
  h, w, _ = image.shape
  if h > w:
    diff = h - w
    l = diff // 2
    r = diff - l
  elif h < w:
    diff = w - h
    t = diff // 2
    b = diff - t
  else:
    return image
  return np.pad(image, ((t, b), (l, r), (0, 0)), constant_values=0)

def square_images(filenames: Sequence[str],
                 out_filenames: Sequence[str] = None):
  for i, fname in enumerate(filenames):
    try:
      arr = np.asarray(PIL.Image.open(fname))
    except PIL.UnidentifiedImageError as uie:
      print(uie)
      print("Skipping non image file")
      continue

    img = PIL.Image.fromarray(square_pad(arr))
    outname = fname if out_filenames is None else out_filenames[i]

    img.save(outname)


@tf.function
def normalize(tensor_image):
  return tf.cast(tensor_image, tf.float32) / 255.0






""" DOC STRING FOR pix2pix GAN Model
"""
import os
import datetime
import numpy as np
import tensorflow as tf
from ganondorf.data import datasets as gds

from .generator import Generator, generator_loss
from .discriminator import Discriminator, discriminator_loss
from .run import fit

def dir_path(dirname: str) -> str:
  return os.path.join(os.path.dirname(__file__), dirname)

def patch_dataset(pre, post):
  pre_patches  = gds.split_into_patches(pre)
  post_patches = gds.split_into_patches(post)

  return (pre_patches, post_patches)

def image_norm(image):
  max_image = np.amax(image, axis=0)
  max_image[max_image == 0] = 1
  return image / max_image

def main():

  btl = gds.BrainTumorProgressionLoader()

  BUFFER_SIZE = 400 #1000
  # BATCH_SIZE = 1
  # IMG_WIDTH = 256
  # IMG_HEIGHT = 256
  EPOCHS = 150  #100,000 in paper and another 100,000 after pretraining :O

  train_pre  = tf.data.Dataset.from_tensor_slices(
      list(map(image_norm, btl.pre_scans[:10]))
      )
  train_post = tf.data.Dataset.from_tensor_slices(
      list(map(image_norm, btl.post_scans[:10]))
      )

  test_pre  = tf.data.Dataset.from_tensor_slices(
      list(map(image_norm, btl.pre_scans[10:]))
      )
  test_post = tf.data.Dataset.from_tensor_slices(
      list(map(image_norm, btl.post_scans[10:]))
      )

  train_dataset = tf.data.Dataset.zip((train_pre, train_post))
  test_dataset  = tf.data.Dataset.zip((test_pre, test_post))

  train_dataset = train_dataset.shuffle(BUFFER_SIZE)
  test_dataset  = test_dataset.shuffle(BUFFER_SIZE)

  #Ignore Batching for now due to patch splitting
  #train_dataset = train_dataset.batch(BATCH_SIZE)
  #test_dataset  = test_dataset.batch(BATCH_SIZE)


  train_dataset = train_dataset.map(patch_dataset)
  test_dataset  = test_dataset.map(patch_dataset)

  generator = Generator()
  discriminator = Discriminator()

  generator_optimizer = tf.keras.optimizers.Adam(learning_rate=0.0001,
                                                 beta_1=0.5)

  discriminator_optimizer = tf.keras.optimizers.Adam(learning_rate=0.0001,
                                                     beta_1=0.5)

  checkpoint_dir = dir_path("training_checkpoints")

  checkpoint_prefix = os.path.join(checkpoint_dir, "ckpt")
  checkpoint = tf.train.Checkpoint(
      generator_optimizer=generator_optimizer,
      discriminator_optimizer=discriminator_optimizer,
      generator=generator,
      discriminator=discriminator)


  log_dir = dir_path("logs")

  summary_writer = tf.summary.create_file_writer(
      os.path.join(log_dir,
                   "fit",
                   datetime.datetime.now().strftime("%Y%m%d-%H;%M;%S"))
      )

  fit(generator=generator,
      generator_loss=generator_loss,
      generator_optimizer=generator_optimizer,
      discriminator=discriminator,
      discriminator_loss=discriminator_loss,
      discriminator_optimizer=discriminator_optimizer,
      train_dataset=train_dataset,
      test_dataset=test_dataset,
      epochs=EPOCHS,
      checkpoint=checkpoint,
      summary_writer=summary_writer,
      checkpoint_prefix=checkpoint_prefix)








""" DOC STRING FOR pix2pix Discriminator Module
"""
import functools
import tensorflow as tf
from ganondorf.layers import encoder_block_3D #, InstNormalize

LeakyReLUMed = functools.partial(tf.keras.layers.LeakyReLU,
                                 alpha=0.2,
                                 )

def Discriminator(): #pylint: disable=C0103
  #initializer = tf.random_normal_initializer(0., 0.02)

  generated = tf.keras.layers.Input(shape=[24,32,32,1], name="input_image")
  real      = tf.keras.layers.Input(shape=[24,32,32,1], name="target_image")

  encoder_stack = [
      encoder_block_3D(32,
                    name="encode_disc_block_1",
                    activation=LeakyReLUMed),  # (bs, 12, 16, 16, 32)
      encoder_block_3D(64,
                    name="encode_disc_block_2",
                    activation=LeakyReLUMed),  # (bs,  6,  8,  8, 64)
      encoder_block_3D(128,
                    name="encode_disc_block_3",
                    activation=LeakyReLUMed),  # (bs,  3,  4,  4, 128)
      ]

  final_conv = tf.keras.layers.Conv3D(filters=32,
                                      kernel_size=(3, 3, 3),
                                      strides=(1, 1, 1),
                                      padding="valid")# ????
  #final_norm = InstNormalize() #?????Do we normalise the final one???

  final_activation = LeakyReLUMed()

  last = tf.keras.layers.Dense(units=1)

  # Functional API run of the model

  x = tf.keras.layers.concatenate([generated, real])

  # Downsample through the model
  for encode in encoder_stack:
    x = encode(x)

  x = final_conv(x)
  x = final_activation(x)

  x = last(x)

  return tf.keras.Model(inputs=[generated, real], outputs=x)



def discriminator_loss(disc_gen_output, disc_real_output):
  """ DOC

      L_dis = E(x,y)~(Pre, Post)[(D(post)(y) − 1)² + D(post)(G(pre->post)(x))²
                               + (D(pre)(x)  − 1)² + D(pre)(G(post->pre)(y))²]

  """

  return tf.reduce_mean(
      (disc_real_output - 1) ** 2 + \
      (disc_gen_output) ** 2
      )

loss_object = tf.keras.losses.BinaryCrossentropy(from_logits=True)

def discriminator_loss_alt(disc_generated_output, disc_real_output):
  real_loss = loss_object(tf.ones_like(disc_real_output), disc_real_output)

  generated_loss = loss_object(
      tf.zeros_like(
          disc_generated_output
          ),
      disc_generated_output
      )

  total_disc_loss = real_loss + generated_loss

  return total_disc_loss


































if __name__ == "__main__":
  BATCH_SIZE = 32
  IMG_SIZE = (160, 160)#(256, 256)

  dataset_path = os.path.join(
      "..","data","datasets","datasets","ALDataset"
      )

  train_dataset = image_dataset_from_directory(
      os.path.join(dataset_path, "train"),
      shuffle=True,
      batch_size=BATCH_SIZE,
      image_size=IMG_SIZE
    )

  test_dataset = image_dataset_from_directory(
      os.path.join(dataset_path, "test"),
      shuffle=True,
      batch_size=BATCH_SIZE,
      image_size=IMG_SIZE
    )

  train_dataset = train_dataset.prefetch(4)
  test_dataset = test_dataset.prefetch(4)

  data_augmentation = tf.keras.Sequential([
      tf.keras.layers.experimental.preprocessing.RandomFlip("horizontal"),
      tf.keras.layers.experimental.preprocessing.RandomRotation(0.2),
      ])

  preprocess_input = tf.keras.applications.mobilenet_v2.preprocess_input

  IMG_SHAPE = IMG_SIZE + (3,)

  base_model = tf.keras.applications.MobileNetV2(input_shape=IMG_SHAPE,
                                                 include_top=False,
                                                 weights="imagenet")

  image_batch, label_batch = next(iter(train_dataset))
  # feature_batch = base_model(image_batch)

  base_model.trainable = False

  global_average_layer = tf.keras.layers.GlobalAveragePooling2D()
  # feature_batch_average = global_average_layer(feature_batch)

  prediction_layer = tf.keras.layers.Dense(1)
  # prediction_batch = prediction_layer(feature_batch_average)

  inputs = tf.keras.Input(shape=(160,160,3))
  x = data_augmentation(inputs)
  x = preprocess_input(x)
  x = base_model(x, training=False)
  x = global_average_layer(x)
  x = tf.keras.layers.Dropout(0.2)(x)
  outputs = prediction_layer(x)
  model = tf.keras.Model(inputs, outputs)

  base_learning_rate = 0.0001
  model.compile(optimizer= \
                tf.keras.optimizers.Adam(learning_rate=base_learning_rate),
                loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
                metrics=['accuracy'])

  print(model.summary())

  initial_epochs = 10
  loss0, accuracy0 = model.evaluate(test_dataset)

  print("initial loss: {:.2f}".format(loss0))
  print("initial accuracy: {:.2f}".format(accuracy0))

  history = model.fit(train_dataset,
                      epochs=initial_epochs,
                      validation_data=test_dataset)

  model.save("no_fine_chkpt/")


  ##################################################################
  # acc = history.history['accuracy']
  # val_acc = history.history['val_accuracy']
  
  # loss = history.history['loss']
  # val_loss = history.history['val_loss']
  
  # plt.figure(figsize=(8, 8))
  # plt.subplot(2, 1, 1)
  # plt.plot(acc, label='Training Accuracy')
  # plt.plot(val_acc, label='Validation Accuracy')
  # plt.legend(loc='lower right')
  # plt.ylabel('Accuracy')
  # plt.ylim([min(plt.ylim()),1])
  # plt.title('Training and Validation Accuracy')
  
  # plt.subplot(2, 1, 2)
  # plt.plot(loss, label='Training Loss')
  # plt.plot(val_loss, label='Validation Loss')
  # plt.legend(loc='upper right')
  # plt.ylabel('Cross Entropy')
  # plt.ylim([0,1.0])
  # plt.title('Training and Validation Loss')
  # plt.xlabel('epoch')
  # plt.show()

  ##################################################################

  fine_tune_at = 100

  base_model.trainable = True

  for layer in base_model.layers[:fine_tune_at]:
    layer.trainable = False


  model.compile(loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
                optimizer = tf.keras.optimizers.RMSprop(
                    learning_rate=base_learning_rate / 10),
                metrics=['accuracy'])

  print(model.summary())

  fine_tune_epochs = 10
  total_epochs = initial_epochs + fine_tune_epochs

  history_fine = model.fit(train_dataset,
                           epochs=total_epochs,
                           initial_epoch=history.epoch[-1],
                           validation_data=test_dataset)

  model.save("fine_chkpt/")

  loss, accuracy = model.evaluate(test_dataset)
  print('loss = {},\naccuracy = {}'.format(loss, accuracy))




import sys
import os
import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import ganondorf as gd
#import ganondorf.pix2pix as pix2pix # residual
#from ganondorf.pix2pix import blocks

example_image = None
example_mask = None

# def nearest_convolve(arr):
#   out = np.empty_like(arr)
#   height = arr.shape[0]
#   width = arr.shape[1]
#   for i in range(height):
#     for j in range(width):


# @tf.function
# def normalize(tensor_image):
#   return tf.cast(tensor_image, tf.float32) / 255.0

@tf.function
def load_image_train(image, mask):

  if tf.random.uniform(()) > 0.5:
    image = tf.image.flip_left_right(image)
    mask  = tf.image.flip_left_right(mask)

  image = gd.data.normalize(image)

  return image, mask

@tf.function
def load_image_test(image, mask):
  input_image = gd.data.normalize(image)
  return input_image, mask
  
def unet_model(output_channels, down_stack, up_stack, residual_stack=None):
  inputs = tf.keras.layers.Input(shape=[*IMAGE_SHAPE,3])

  skips = down_stack(inputs)
  x = skips[-1]
  skips = reversed(skips)

  if residual_stack is not None:
    for res in residual_stack:
      x = res(x)

  for up, skip in zip(up_stack, skips):
    concat = tf.keras.layers.Concatenate()
    x = concat([x, skip])
    x = up(x)

  last = tf.keras.layers.Conv2D(output_channels,
                                3,
                                strides=1,
                                padding='same',
                                activation="tanh")
  # x = pen(x)
  x = last(x)

  return tf.keras.Model(inputs=inputs, outputs=x)

@tf.function
def create_mask(pred_mask):
  pred_mask = tf.argmax(pred_mask, axis=-1)
  pred_mask = pred_mask[..., tf.newaxis]
  return pred_mask[0]

class DisplayCallback(tf.keras.callbacks.Callback):
  def on_epoch_end(self, epoch, logs=None):
    #clear_output(wait=True)
    if (epoch + 1) % 10 == 0:
      gd.data.Visualizer.show_image_predictions(
          self.model,
          dataset=(example_image, example_mask),
          format_prediction=create_mask
          )
    print('\nSample Predictions after epoch {}\n'.format(epoch+1))


if __name__ == '__main__':
  EPOCHS = 100
  OUTPUT_CHANNELS = 2
  IMAGE_SHAPE = (128, 128)

  TRAIN_LENGTH = 34
  BATCH_SIZE = 8
  BUFFER_SIZE = 16 #1000
  STEPS_PER_EPOCH = 1
  VAL_SUBSPLITS = 5
  VALIDATION_STEPS = 1

  train_dataset, test_dataset = gd.data.Dataset.load("ALRing", #"ALSegmentation",
                                                     size=IMAGE_SHAPE)
  train_dataset = train_dataset.map(load_image_train,
                                    num_parallel_calls=tf.data.AUTOTUNE)

  test_dataset = test_dataset.map(load_image_test)

  train_dataset = train_dataset.cache() \
                               .shuffle(BUFFER_SIZE) \
                               .batch(BATCH_SIZE).repeat()
  train_dataset = train_dataset.prefetch(buffer_size=tf.data.AUTOTUNE)

  example_image, example_mask  = next(iter(test_dataset))

  test_dataset = test_dataset.batch(BATCH_SIZE)


  base_model = tf.keras.applications.MobileNetV2(input_shape=[*IMAGE_SHAPE, 3],
                                                 include_top=False)
  layer_names = [
      'block_1_expand_relu',  #  64x64x96
      'block_3_expand_relu',  #  32x32x144
      'block_6_expand_relu',  #  16x16x192
      'block_13_expand_relu', #  8x8x576
      # 'block_16_project',     #  4x4x320
      'Conv_1', # 4x4x1280
      ]

  base_model_outputs = \
      [base_model.get_layer(name).output for name in layer_names]

  down_stack = tf.keras.Model(inputs=base_model.input,
                              outputs=base_model_outputs)

  down_stack.trainable = False

  residual_stack = [
      gd.layers.ResidualBottleneckLayer.as_residual_bridge(2, 1280), # 320),
      gd.layers.ResidualBottleneckLayer.as_residual_bridge(2, 1280), # 320),
      gd.layers.ResidualBottleneckLayer.as_residual_bridge(2, 1280), # 320),
      tf.keras.layers.ReLU(),
      ]
  
  up_stack = [
      gd.layers.decoder_block_2D(1280, name="decode_block_1"), #512
      gd.layers.decoder_block_2D(576, name="decode_block_2"), #256
      gd.layers.decoder_block_2D(192, name="decode_block_3"), #128
      gd.layers.decoder_block_2D(144,  name="decode_block_4"), #64
      gd.layers.decoder_block_2D(96,  name="decode_block_5"), #64
      ]


  model = unet_model(OUTPUT_CHANNELS,
                     down_stack, up_stack, residual_stack=residual_stack)

  model.compile(optimizer='adam',
                loss=tf.keras.losses.SparseCategoricalCrossentropy(
                    from_logits=True
                    ),
                metrics=['accuracy'])

  tf.keras.utils.plot_model(model, show_shapes=True, dpi=64, to_file="plot.svg")

  ring_sample_weights = gd.data.Dataset.get_sample_weights_func([3.075, 1])

  model_history = model.fit(train_dataset.map(ring_sample_weights),
                            epochs=EPOCHS,
                            steps_per_epoch=STEPS_PER_EPOCH,
                            validation_steps=VALIDATION_STEPS,
                            validation_data=test_dataset,
                            callbacks=[DisplayCallback()])


  gd.data.Visualizer.show_image_predictions(
      model,
      dataset=(example_image, example_mask),
      format_prediction=create_mask
      )

  loss = model_history.history['loss']
  val_loss = model_history.history['val_loss']

  plt.figure()
  plt.plot(model_history.epoch, loss, 'r', label='Training loss')
  plt.plot(model_history.epoch, val_loss, 'bo', label='Validation loss')
  plt.title('Training and Validation Loss')
  plt.xlabel('Epoch')
  plt.ylabel('Loss Value')
  plt.ylim([0, 1])
  plt.legend()
  plt.show()

  val_loss_arr = np.array(val_loss)
  val_acc_arr = np.array(model_history.history['val_accuracy'])
  
  val_acc_best_index = np.argmax(val_acc_arr)
  val_loss_best_index = np.argmin(val_loss_arr)
  
  print("Highest Accuracy Epoch:", val_acc_best_index + 1,
        "Lowest Loss Epoch:", val_loss_best_index + 1)
  
  print("Highest Accuracy:", val_acc_arr[val_acc_best_index],
        "Lowest Loss:", val_loss_arr[val_loss_best_index])

  if val_loss_best_index != val_acc_best_index:
    print("Loss and acc not at same location")
    print("Loss Accuracy:", val_acc_arr[val_loss_best_index],
          "Accuracy Loss:", val_loss_arr[val_acc_best_index])

  if len(sys.argv) < 2 or sys.argv[1].lower() != "nosave":
    # model.save("chkpt/")
    model.save("E{:02}_chkpt/".format(EPOCHS))





""" Dataset class module docstring

"""
from typing import ClassVar, Callable, Union, TypeVar, Mapping, Sequence

import numpy as np
import tensorflow as tf
import ganondorf.core.datacore as datacore
import ganondorf.data.dataset_loader as dl

T = TypeVar("T")
LoadedDataset = Union[tf.data.Dataset, tuple]
DatasetLoaderFunc = Callable[..., LoadedDataset]

def placeholder(*args, **kwargs):
  return (0,)

class Dataset(datacore.Dataset):

  _DATASET_LOADER_MAP: ClassVar[dict[str, DatasetLoaderFunc]] = \
      {# "arcm"
       "arcm": dl.load_arcm,
       # "har"
       "har": dl.load_har,
       # "mHealth"
       "mhealth": dl.load_mhealth,
       # "BrainTumorProgression"
       "braintumorprogression": dl.load_brain_tumor_progression,
       # "BrainLesionsGlial"
       "brainlesionsglial": placeholder,
       # "ALIntraoperative"
       "alintraoperative": placeholder,
       # "ALSegmentation"
       "alsegmentation": dl.load_AL_segmentation,
       # "ALRing"
       "alring": dl.load_AL_ring,
       }

  @staticmethod
  def load(dataset_name:str, *args, **kwargs)->LoadedDataset: #normalize=False
    return Dataset._DATASET_LOADER_MAP[dataset_name.lower()](*args, **kwargs)


  @staticmethod
  def get_sample_weights_func(weights:Sequence[float]):
    def sample_weights_func(inputs, targets):
      class_weights = tf.constant(weights)
      class_weights = class_weights / tf.reduce_sum(class_weights)
  
      sample_weights = tf.gather(class_weights, 
                                 indices=tf.cast(targets, tf.int32))
  
      return inputs, targets, sample_weights
  
    return tf.function(sample_weights_func)    





""" ImageDataset class module docstring

"""

import tensorflow as tf
from . import dataset

class ImageDataset(dataset.Dataset):

  @staticmethod
  @tf.function
  def normalize_image(tensor_image):
    return tf.cast(tensor_image, tf.float32) / 255.0






""" Data Package Docstring

"""

from .dataset import Dataset
from .visualizer import Visualizer
from .format_image_data import *

# from . import format_image_data as fid

# __all__ = ['Dataset', 'Visualizer']
# __all__ += fid.__all__




""" DOC STRING FOR pix2pix Generator Module
"""
import tensorflow as tf
from ganondorf.layers import ResidualBottleneckLayer
from ganondorf.layers import encoder_block_3D, decoder_block_3D, Feature3D

def Generator(): # pylint: disable=C0103
  OUTPUT_CHANNELS = 1 # pylint: disable=C0103

  inputs = tf.keras.layers.Input(shape=[24,32,32,1])

  encoder_stack = [
      encoder_block_3D(32,
                    name="encode_block_1",
                    activation=tf.keras.layers.ReLU),  # (bs, 12, 16, 16, 32)
      encoder_block_3D(64,
                    name="encode_block_2",
                    activation=tf.keras.layers.ReLU),  # (bs,  6,  8,  8, 64)
      encoder_block_3D(128,
                    name="encode_block_3",
                    activation=tf.keras.layers.ReLU),  # (bs,  3,  4,  4, 128)
      ]

  residual_stack = [
      ResidualBottleneckLayer.as_residual_bridge(3, 128),
      ResidualBottleneckLayer.as_residual_bridge(3, 128),
      ResidualBottleneckLayer.as_residual_bridge(3, 128),
      ]

  after_residual_activation = tf.keras.layers.ReLU()

  decoder_stack = [
      decoder_block_3D(128, name="decode_block_1"),  # (bs,  6,  8,  8, 128)
      decoder_block_3D(64,  name="decode_block_2"),  # (bs, 12, 16, 16, 64)
      decoder_block_3D(32,  name="decode_block_3"),  # (bs, 24, 32, 32, 32)
      ]


  initializer = tf.random_normal_initializer(0., 0.02)
  last = Feature3D(filters=OUTPUT_CHANNELS,
                 strides=1,
                 kernel_initializer=initializer,
                 activation="tanh")  # (bs, 24, 32, 32, 1)

  x = inputs

  # Downsample through the model
  skips = []
  for encode in encoder_stack:
    x = encode(x)
    skips.append(x)

  skips.reverse()

  # Residual blocks through the model
  for res_block in residual_stack:
    x = res_block(x)

  x = after_residual_activation(x)

  # Upsample and establish skips
  for decode, skip in zip(decoder_stack, skips):
    x = tf.keras.layers.Concatenate()([x, skip])  # ????? or above?
    x = decode(x)

  x = last(x)

  return tf.keras.Model(inputs=inputs, outputs=x)

def generator_loss(target, gen_output, disc_gen_output):
  """ DOC

      L_cycle = Not Implemented
        E(x,y)~(Pre_pair, Post_pair)(||G(pre -> post)[G(post -> pre)(x)] - x||₁)

      pair_loss = E(x,y)~(Pre_pair, Post_pair)(||G(pre -> post)(x) − y||₁

      L_adv = E(x,y)~(Pre, Post)[(D(post)(G(pre -> post)(x)) − 1)²]
  """

  pair_loss = tf.reduce_mean(tf.abs(gen_output - target))

  adversarial_loss = tf.reduce_mean((disc_gen_output - 1) ** 2)

  # approx of lambda2 from paper redistributed to 3%
  total_loss = pair_loss + 0.03 * adversarial_loss

  return total_loss, pair_loss, adversarial_loss, None


loss_object = tf.keras.losses.BinaryCrossentropy(from_logits=True)


def generator_loss_alt(target, gen_output, disc_generated_output):
  gan_loss = loss_object(
      tf.ones_like(
          disc_generated_output
          ),
      disc_generated_output
      )

  # mean absolute error
  l1_loss = tf.reduce_mean(tf.abs(target - gen_output))

  total_gen_loss = gan_loss + (100 * l1_loss)

  return total_gen_loss, gan_loss, l1_loss, None















""" Runner Docstring

"""
import os
import time
import tensorflow as tf
import matplotlib.pyplot as plt
import ganondorf.data.datasets as gds

def show_model_output(model, input_data):

  output = model(input_data, training=False)

  plt.subplot(121)
  plt.imshow(input_data[0, 0, ...], cmap="Blues")

  plt.subplot(122)
  plt.imshow(output[0, 0, ...], cmap="Blues")

  plt.show()

def plot_model(model, savename: str = None):
  if savename is None:
    savename = f"{model.name}.png" if model.name != "" else "model_plot.png"

  tf.keras.utils.plot_model(model, show_shapes=True, dpi=64, to_file=savename)

def generate_images(model, test_input, target, epoch:int):
  prediction = model(test_input, training=True)
  plt.figure(figsize=(12, 12))

  test_input = gds.sew_patches(test_input)
  target = gds.sew_patches(target)
  prediction = gds.sew_patches(prediction)

  display_list = [test_input[15], target[15], prediction[15]]
  title = ["Input Image", "Ground Truth", "Predicted Image"]

  for i in range(3):
    plt.subplot(1, 3, i+1)
    plt.title(title[i])

    plt.imshow(display_list[i])  # May need to window and level
    plt.axis("off")

  plt.savefig(os.path.join(os.path.dirname(__file__),
                           "..",
                           "tensorflow_outputs",
                           "Generator_Output",
                           "{:02}_Generated_Output.jpg".format(epoch)),
              dpi=96)


@tf.function
def train_step_tf_tut(input_image,
               target,
               generator,
               generator_loss,
               generator_optimizer,
               discriminator,
               discriminator_loss,
               discriminator_optimizer,
               epoch,
               summary_writer=None):

  with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:
    gen_output = generator(input_image, training=True)

    disc_real_output = discriminator([input_image, target], training=True)
    disc_generated_output = discriminator([input_image, gen_output],
                                          training=True)

    gen_loss, pair_loss, adversarial_loss, _ = generator_loss(target,
                                                        gen_output,
                                                        disc_generated_output)
    disc_loss = discriminator_loss(disc_generated_output,
                                   disc_real_output)

  generator_gradients = gen_tape.gradient(
      gen_loss,
      generator.trainable_variables)

  discriminator_gradients = disc_tape.gradient(
      disc_loss,
      discriminator.trainable_variables)


  generator_optimizer.apply_gradients(
      zip(generator_gradients,
          generator.trainable_variables)
      )

  discriminator_optimizer.apply_gradients(
      zip(discriminator_gradients,
          discriminator.trainable_variables)
      )


  if summary_writer is not None:
    with summary_writer.as_default():
      tf.summary.scalar("gen_total_loss", gen_loss, step=epoch)
      tf.summary.scalar("pair_loss", pair_loss, step=epoch)
      tf.summary.scalar("adversarial_loss", adversarial_loss, step=epoch)
      tf.summary.scalar("disc_loss", disc_loss, step=epoch)





#We adopted Adam [38] as the optimizer to train the model
#with a learning rate of 10−4. Other hyper-parameters were
#empirically set as K = 3, α = 1, λ1 = 0.5, λ2 = 0.05,
#λ3 = 0.1, λ4 = 0.5, and λ5 = 0.5.




@tf.function
def train_step(input_image,
               target,
               generator,
               generator_loss,
               generator_optimizer,
               discriminator,
               discriminator_loss,
               discriminator_optimizer,
               epoch,
               summary_writer=None):

  # Train Discriminator with fixed Generator
  with tf.GradientTape() as disc_tape:
    gen_output = generator(input_image, training=True)

    disc_real_output = discriminator([input_image, target], training=True)
    disc_generated_output = discriminator([input_image, gen_output],
                                          training=True)

    disc_loss = discriminator_loss(disc_generated_output,
                                   disc_real_output)

  discriminator_gradients = disc_tape.gradient(
      disc_loss,
      discriminator.trainable_variables)


  discriminator_optimizer.apply_gradients(
      zip(discriminator_gradients,
          discriminator.trainable_variables)
      )

  with tf.GradientTape() as gen_tape:
    gen_output = generator(input_image, training=True)

    disc_real_output = discriminator([input_image, target], training=True)
    disc_generated_output = discriminator([input_image, gen_output],
                                          training=True)

    gen_loss, pair_loss, adversarial_loss, _ = generator_loss(target,
                                                        gen_output,
                                                        disc_generated_output)
  generator_gradients = gen_tape.gradient(
      gen_loss,
      generator.trainable_variables)

  generator_optimizer.apply_gradients(
      zip(generator_gradients,
          generator.trainable_variables)
      )

  if summary_writer is not None:
    with summary_writer.as_default():
      tf.summary.scalar("gen_total_loss", gen_loss, step=epoch)
      tf.summary.scalar("pair_loss", pair_loss, step=epoch)
      tf.summary.scalar("adversarial_loss", adversarial_loss, step=epoch)
      tf.summary.scalar("disc_loss", disc_loss, step=epoch)





def fit(generator,
        generator_loss,
        generator_optimizer,
        discriminator,
        discriminator_loss,
        discriminator_optimizer,
        train_dataset,
        test_dataset,
        epochs,
        checkpoint,
        batch_size:int = 16, #32 is too large and 8 is a close second
        summary_writer=None,
        checkpoint_prefix="ckpt"):

  for epoch in range(epochs):
    start = time.time()

    for example_input, example_target in test_dataset.take(1):
      generate_images(generator, example_input, example_target, epoch)
    print("Epoch: ", epoch)

    # Train
    for n, (input_image, target) in train_dataset.enumerate():
      print(".", end="")
      if (n + 1) % 100 == 0:
        print()

      for batch_start in range(0, len(input_image), batch_size):
        input_image_patch = input_image[batch_start : batch_start + batch_size]
        target_patch = target[batch_start : batch_start + batch_size]

        train_step(input_image_patch,
                   target_patch,
                   generator,
                   generator_loss,
                   generator_optimizer,
                   discriminator,
                   discriminator_loss,
                   discriminator_optimizer,
                   epoch,
                   summary_writer=summary_writer)
    print()

    # saving (checkpoint) the model every 20 epochs
    if (epoch + 1) % 20 == 0:
      checkpoint.save(file_prefix=checkpoint_prefix)

    print(f"Time taken for epoch {epoch + 1} is {time.time() - start} sec\n")

  checkpoint.save(file_prefix=checkpoint_prefix)











import sys

import PIL
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt

import ganondorf.data



@tf.function
def normalize(tensor_image):
  return tf.cast(tensor_image, tf.float32) / 255.0


@tf.function
def load_image_test(image, mask):
  input_image = normalize(image)
  return input_image, mask

@tf.function
def create_mask(pred_mask):
  pred_mask = tf.argmax(pred_mask, axis=-1)
  pred_mask = pred_mask[..., tf.newaxis]
  return pred_mask[0]



if __name__ == '__main__':
  if len(sys.argv) < 3:
    print("Requires path to image being passed in."
          " and path to model too."
          " Exiting ....")

  model = tf.keras.models.load_model(sys.argv[2])

  if sys.argv[1].lower() == "all":
    _, test_dataset = ganondorf.data.Dataset.load("ALSegmentation",
                                                  size=(128,128))
    test_dataset = test_dataset.map(load_image_test)    
    ganondorf.data.Visualizer.show_image_predictions(
      model,
      test_dataset.batch(1),
      num=20,
      format_prediction=create_mask)

  else:

    img = PIL.Image.open(sys.argv[1]).convert('RGB').resize((128,128))
    arr = np.asarray(img)[np.newaxis,...]


    arr = arr.astype(np.float32) / 255.0

    predict = create_mask(model.predict(arr))
    plt.imshow(predict)
    plt.show()
    #----------------------------------------
    predict = predict.numpy()
    predict = predict.astype(np.uint8)
    predict = predict.reshape(predict.shape[0], predict.shape[1])
    output = PIL.Image.fromarray(predict * 255)
    output = output.convert('1')
    output.save('prediction.png')







from typing import Union, Callable, Any, Sequence

import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt

# May not need be array could be any??
Datapoint = Union[tf.data.Dataset, tuple[np.array, np.array]]

identity = lambda x: x

class Visualizer():

  @staticmethod
  def display_image(*display_list, title:Sequence=None):
    #plt.figure(figsize=(15,15))
    if title is None:
      title = range(1, len(display_list) + 1)

    for i in range(len(display_list)):
      plt.subplot(1, len(display_list), i+1)
      plt.title(title[i])
      plt.imshow(display_list[i])
      plt.axis('off')
    plt.show()

#
  @staticmethod
  def display_image_predictions(image, label, prediction):
    Visualizer.display_image(image,
                             label,
                             prediction,
                             title = ['Input', 'Labeled', 'Predicted'])

  @staticmethod
  def show_image_predictions(
      model:tf.keras.Model,
      dataset:Datapoint=None,
      num:int=1,
      format_prediction:Callable[np.array, np.array]=identity
      )->None:

    if isinstance(dataset, tf.data.Dataset):
      for image, label in dataset.take(num):
        print("okay")
        prediction = model.predict(image)
        Visualizer.display_image_predictions(image[0], label[0],
                                             format_prediction(prediction))
    else:
      prediction = model.predict(dataset[0][np.newaxis, ...])
      Visualizer.display_image_predictions(dataset[0], dataset[1],
                                           format_prediction(prediction))  



""" DOC STRING FOR pix2pix Blocks Module

Block are reusable sequences of layers
"""
from typing import Callable
import functools
import tensorflow as tf
import tensorflow_addons as tfa

__all__ = ['Feature2D', 'Feature3D', 'Decode2D', 'Decode3D', 'InstNormalize',
           'encoder_block', 'encoder_block_3D', 'encoder_block_2D',
           'decoder_block', 'decoder_block_3D', 'decoder_block_2D']


Feature3D = functools.partial(tf.keras.layers.Conv3D,
                              kernel_size=(3,3,3),
                              padding="same",
                              #use_bias=False
                              )

Decode3D = functools.partial(tf.keras.layers.Conv3DTranspose,
                             kernel_size=(3,3,3),
                             padding="same",
                             use_bias=False,
                             )

Feature2D = functools.partial(tf.keras.layers.Conv2D,
                              kernel_size=(3,3),
                              padding="same",
                              #use_bias=False
                              )

Decode2D = functools.partial(tf.keras.layers.Conv2DTranspose,
                             kernel_size=(3,3),
                             padding="same",
                             use_bias=False,
                             )

InstNormalize = functools.partial(tfa.layers.InstanceNormalization,
                                  axis=-1, center=True, scale=True,
                                  beta_initializer="random_uniform",
                                  gamma_initializer="random_uniform"
                                  )

def encoder_block(
    filters: int,
    feature_layer:tf.keras.layers.Layer,
    name: str="",
    instance_norm_mask: int = 0b11,
    activation: Callable[[], tf.keras.layers.Layer] = tf.keras.layers.ReLU
    ) -> tf.keras.Sequential:
  """ Encoder Block Creator function

  Parameters
  ----------

  filters : int
    Number of filters to use for the block"s two convolutional layers.

  name : str, optional
    The name to use for the block to aid in debugging, defaults to "encoder".

  apply_instancenorm : int, default 0b11
    Binary mask to decide where to apply instance normalisation, the bit
    location defines whether or not to apply instance normalisation after
    the corresponding layer, i.e. 0b10 means apply instance normalisation
    after layer 2 but not after layer 1.

  """
  initializer = tf.random_normal_initializer(0.0, 0.02)

  result = tf.keras.Sequential(name=name)

  result.add(feature_layer(filters, strides=1, kernel_initializer=initializer))

  if instance_norm_mask & 0b01:
    result.add(InstNormalize())

  if activation is not None:
    result.add(activation())

  result.add(feature_layer(filters, strides=2, kernel_initializer=initializer))

  if instance_norm_mask & 0b10:
    result.add(InstNormalize())

  if activation is not None:
    result.add(activation())

  return result


def decoder_block(
    filters: int,
    feature_layer:tf.keras.layers.Layer,
    decode_layer:tf.keras.layers.Layer,
    name: str="",
    instance_norm_mask: int = 0b11,
    activation: Callable[[], tf.keras.layers.Layer] = tf.keras.layers.ReLU
    ) -> tf.keras.Sequential:
  """ Decoder Block Creator function

  Parameters
  ----------

  filters : int
    Number of filters to use for the block's convolutional and
    transposed convolutional layers.

  name : str, optional
    The name to use for the block to aid in debugging, defaults to "decoder".

  apply_instancenorm : int, default 0b11
    Binary mask to decide where to apply instance normalisation, the bit
    location defines whether or not to apply instance normalisation after
    the corresponding layer, i.e. 0b10 means apply instance normalisation
    after layer 2 but not after layer 1.

  """
  initializer = tf.random_normal_initializer(0.0, 0.02)

  result = tf.keras.Sequential(name=name)

  result.add(decode_layer(filters, strides=2, kernel_initializer=initializer))

  if instance_norm_mask & 0b01:
    result.add(InstNormalize())

  if activation is not None:
    result.add(activation())

  result.add(feature_layer(filters, strides=1, kernel_initializer=initializer))

  if instance_norm_mask & 0b10:
    result.add(InstNormalize())

  if activation is not None:
    result.add(activation())

  return result

def encoder_block_3D(
    filters: int,
    name: str="",
    instance_norm_mask: int = 0b11,
    activation: Callable[[], tf.keras.layers.Layer] = tf.keras.layers.ReLU
    ) -> tf.keras.Sequential:

  return encoder_block(filters, Feature3D, name, instance_norm_mask, activation)

def decoder_block_3D(
    filters: int,
    name: str="",
    instance_norm_mask: int = 0b11,
    activation: Callable[[], tf.keras.layers.Layer] = tf.keras.layers.ReLU
    ) -> tf.keras.Sequential:

  return decoder_block(filters, 
                       Feature3D, Decode3D,
                       name, instance_norm_mask,
                       activation)

def encoder_block_2D(
    filters: int,
    name: str="",
    instance_norm_mask: int = 0b11,
    activation: Callable[[], tf.keras.layers.Layer] = tf.keras.layers.ReLU
    ) -> tf.keras.Sequential:

  return encoder_block(filters, Feature2D, name, instance_norm_mask, activation)

def decoder_block_2D(
    filters: int,
    name: str="",
    instance_norm_mask: int = 0b11,
    activation: Callable[[], tf.keras.layers.Layer] = tf.keras.layers.ReLU
    ) -> tf.keras.Sequential:

  return decoder_block(filters, 
                       Feature2D, Decode2D,
                       name, instance_norm_mask,
                       activation)










""" Module for loading datasets.
Includes dataset transformations for data of types such as:
  - Images including:
    - Standard iamge formats such as Png's and Jpeg's
    - Medical images such as those saved in medical formats such as .nii (NIfTI)
  
  - Raw data saved in numpy format
"""

#stdlib imports
import os
import importlib.resources as resources
# import collections

#external imports
import PIL
import numpy as np
import tensorflow as tf
import SimpleITK as sitk
import ptils

#intra-package imports
import ganondorf.data.format_image_data as fid
import ganondorf.data.datasets as gdds

IMAGE = 0
MASK = 1

def image_as_array(path: str, mode: str=None) -> np.array:
  image = PIL.Image.open(path)
  return fid.image_as_array(image, mode)

def medical_as_array(path: str) -> np.array:
  image = sitk.ReadImage(path)
  return fid.medical_as_array(image)

def dir_to_mask(path):
  image_names = os.listdir(path)
  for x in image_names:
    img = PIL.Image.open(x)
    img = fid.as_mask(img)
    img.save(x)

def load_image_pair_dataset(path:str, 
                            image_dirname:str="image",
                            label_dirname:str="mask",
                            size:tuple[int, int]=None)->tf.data.Dataset:

  dataset_path = resources.files(gdds).joinpath(path) #.open()

  image_files = dataset_path.joinpath(image_dirname).iterdir()
  label_files = dataset_path.joinpath(label_dirname).iterdir()

  images = []
  labels = []

  for img, lab in zip(image_files, label_files):

    if size is not None:
      image = fid.image_as_array(fid.resize_image(img, size))
      label = fid.image_as_array(fid.resize_image(lab, size), mode='1')
    else:
      #TODO: decide if acceptable or not!
      image = image_as_array(img) # Using path but in zip so not safe? 
      label = image_as_array(lab, mode='1') #  Should already be in this form
       

    # dataset.append( {'image': image, 'label': label} )
    images.append(image)
    labels.append(label)
  
  return tf.data.Dataset.from_tensor_slices((images, labels))

def load_medical_dataset(path:str,
                         data_prefix:str = "",
                         data_suffix:str = ""
                         )->tf.data.Dataset:

  dataset_path = resources.files(gdds).joinpath(path)
  patients = (patient for patient in dataset_path.iterdir() \
              if patient.is_dir() and patient.name != ".git")

  dataset = []

  for patient in patients:
    patient_data = [data for data in patient.iterdir() \
                    if data.startswith(data_prefix) and \
                       data.endswith(data_suffix) ]
    dataset.extend(patient_data)

  return tf.data.Dataset.from_tensor_slices(dataset)

def load_npz_dataset(path:str, label_index:int=-1)->tf.data.Dataset:

  data = resources.files(gdds).joinpath(path)

  dataset = {}

  with np.load(data) as datadict:
    for name in datadict:
      arr, label = np.hsplit(datadict[name], np.array([label_index]))
      dataset[name] = (arr, label)
  
  # return tf.data.Dataset.from_tensor_slices(
  #     collections.OrderedDict(
  #         x = input_array,
  #         y = label_array
  #         )
  #     )

  return tf.data.Dataset.from_tensor_slices(dataset)


def load_segmentation_dataset(path:str,
                              load_train = True,
                              load_test = True,
                              size:tuple[int, int]=None)->tf.data.Dataset:
  dataset = []

  if load_train:
    dataset.append(load_image_pair_dataset(path + "/train",
                                           image_dirname="image",
                                           label_dirname="mask",
                                           size=size))
  if load_test:
    dataset.append(load_image_pair_dataset(path + "/test",
                                           image_dirname="image",
                                           label_dirname="mask",
                                           size=size))
  if dataset == []:
    return None
  elif len(dataset) == 1:
    return dataset[0]
  else:
    return tuple(dataset)


def load_AL_segmentation(load_train = True,
                         load_test = True,
                         size:tuple[int, int]=None)->tf.data.Dataset:
  path = "ALSegment"
  return load_segmentation_dataset(path, load_train, load_test, size)


def load_AL_ring(load_train = True,
                 load_test = True,
                 size:tuple[int, int]=None)->tf.data.Dataset:
  path = "ALRing"
  return load_segmentation_dataset(path, load_train, load_test, size)


def load_brain_tumor_progression(load_pre_scan=True,
                                 load_post_scan=True,
                                 load_pre_mask=True,
                                 load_post_mask=True):

  path = "BrainTumorProgression"

  dataset = []

  if load_pre_scan:
    dataset.append(load_medical_dataset(path, 
                                        data_prefix="Pre_Scan_T1.nii.gz"))

  if load_post_scan:
    dataset.append(load_medical_dataset(path, 
                                        data_prefix="Post_Scan_T1.nii.gz"))

  if load_pre_mask:
    dataset.append(load_medical_dataset(path, 
                                        data_prefix="Pre_Mask.nii.gz"))

  if load_post_mask:
    dataset.append(load_medical_dataset(path, 
                                        data_prefix="Post_Mask.nii.gz"))

  if dataset == []:
    return None
  else:
    return tf.data.Dataset.zip(tuple(dataset))

def load_arcm(normalised:bool=False)->tf.data.Dataset:

  path = ("arcm/"
          "arcm_data_norm.npz" if normalised else "arcm_data.npz")

  return load_npz_dataset(path)

def load_mhealth(normalised:bool=False)->tf.data.Dataset:

  path = ("mHealth/"
          "mHealth_data_norm.npz" if normalised else "mHealth_data.npz")

  return load_npz_dataset(path)

def load_har(normalised:bool=False)->tf.data.Dataset:

  path = ("har/"
          "har_data_norm.npz" if normalised else "har_data.npz")

  return load_npz_dataset(path)






""" Residual module for definition of Residual layers

"""

from typing import Union, ClassVar
import functools
import tensorflow as tf

__all__ = ['ResidualBottleneckLayer']

ConvSpec = Union[int, tuple[int, ...]]

class ResidualBottleneckLayer(tf.keras.Model):
  """ Class for Residual BottleneckLayer

  """

  DOWN_FILTER_SCALE: ClassVar[int] = 4
  _NAME_ACC = 0


  def __init__(self,
               dimensions: int,
               filters: ConvSpec,
               kernel_size: ConvSpec = 3, #(3, 3, 3),
               strides: ConvSpec = 1, #(1, 1, 1),
               name: str = None):
  
    self.feature_layer = None

    if dimensions == 1:
      self.feature_layer = tf.keras.layers.Conv1D
    elif dimensions == 2:
      self.feature_layer = tf.keras.layers.Conv2D
    elif dimensions == 3:
      self.feature_layer = tf.keras.layers.Conv3D
    else:
      raise ValueError("'dimensions' must be either 1, 2 or 3 but was"
                       f"{dimensions}")

    if name is None:
      ResidualBottleneckLayer._NAME_ACC += 1
      name = f"Residual_Bottleneck_{ResidualBottleneckLayer._NAME_ACC}"

    super().__init__(name=name)
    self._filter1, self._filter2, self._filter3 = [filters] * 3 \
        if isinstance(filters, int) else filters

    self._kernel_size = [kernel_size] * 3 \
        if isinstance(kernel_size, int) else kernel_size

    self._strides = [strides] * 3 \
        if isinstance(strides, int) else strides

    # Feature Down Convolution
    self.down_norm = tf.keras.layers.BatchNormalization()
    self.down_act  = tf.keras.layers.ReLU()
    self.down_conv = self.feature_layer(filters=self._filter1,
                                        kernel_size=1,
                                        strides=strides,
                                        padding="valid")

    # kernel_size Convolution
    self.norm = tf.keras.layers.BatchNormalization()
    self.act  = tf.keras.layers.ReLU()
    self.conv = self.feature_layer(filters=self._filter2,
                                   kernel_size=kernel_size,
                                   strides=1,
                                   padding="same")

    # Feature Up Convolution
    self.up_norm = tf.keras.layers.BatchNormalization()
    self.up_act  = tf.keras.layers.ReLU()
    self.up_conv = self.feature_layer(filters=self._filter3,
                                      kernel_size=1,
                                      strides=1,
                                      padding="valid")


  #TODO add insert code for first def after __init__
  def call(self, input_tensor, training=False):
    """ For the first Residual Unit
    (that follows a stand-alone convolutional layer,
    conv1), we adopt the first activation right after conv1
    and before splitting into two paths; for the last Residual Unit
    (followed by average pooling and a fullyconnected classifier),
    we adopt an extra activation right after its element-wise addition

    """

    x = self.down_norm(input_tensor, training=training)
    x = self.down_act(x)
    x = self.down_conv(x)

    x = self.norm(x, training=training)
    x = self.act(x)
    x = self.conv(x)

    x = self.up_norm(x, training=training)
    x = self.up_act(x)
    x = self.up_conv(x)

    x += input_tensor

    return x


  #TODO: fix indenting for this too
  @classmethod
  def as_residual_bridge(cls,
                         dimensions: int,
                         shape: Union[int, tuple[int, int]]):
    """ Doc String

    input -> [(input_shape -> 1/4) -> (1/4 -> 1/4) -> [1/4 -> output_shape)]

    """
    out_shape = shape if isinstance(shape, int) else shape[1]
    scale = out_shape // cls.DOWN_FILTER_SCALE
    filters = [scale, scale, out_shape]

    return cls(dimensions=dimensions, filters=filters, kernel_size=3, strides=1)

  @staticmethod
  def _tuple_reducer(acc, elem) -> str:
    if acc == "":
      return f"({elem})"
    else:
      return f"{acc[:-1]}x{elem})"

  #TODO: Spacing properly i.e. using maths
  def __str__(self):
    kernel_string = functools.reduce(self._tuple_reducer, self._kernel_size, "")

    stride_string = functools.reduce(self._tuple_reducer, self._strides, "")

    return (f"Layer  |  filter->filter |  kernel size |  stride  | Padding\n\n"

            f"Conv 1 |  (input -> {self._filter1})  | "
            f"   (1x1x1)   | {stride_string}  | \"valid\"\n"

            "\t\t\u21D3\n"

            f"Conv 2 |   ({self._filter1} -> {self._filter2})    | "
            f"   {kernel_string}   | (1x1x1)  | \"same\"\n"

            "\t\t\u21D3\n"

            f"Conv 3 |    ({self._filter2} -> {self._filter3})  | "
            f"   (1x1x1)   | (1x1x1)  | \"valid\""
            )




""" Transfer Learning Module DOC STRING

"""

import os
from typing import Tuple

import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import tensorflow as tf
import tensorflow_hub as hub
from tensorflow.keras.preprocessing import image_dataset_from_directory

def predict(model, input_batch, labels=None):

  predictions = model.predict_on_batch(input_batch).flatten()

  # Apply a sigmoid since our model returns logits
  predictions = tf.nn.sigmoid(predictions)
  predictions = tf.where(predictions < 0.5, 0, 1)

  print('Predictions:\n', predictions.numpy())
  if labels is not None:
    print('Labels:\n', labels)

  # plt.figure(figsize=(10, 10))
  # for i in range(len(predictions)):
  #   ax = plt.subplot(3, 3, i + 1)
  #   plt.imshow(image_batch[i].astype("uint8"))
  #   plt.title(class_names[predictions[i]])
  #   plt.axis("off")


if __name__ == "__main__":

  BATCH_SIZE = 32
  IMG_SIZE = (512, 512)

  dataset_path = os.path.join(
      "..","data","datasets","datasets","ALDataset"
      )

  img_path = os.path.join(dataset_path, "Leak", "image_02.png")
  image = np.asarray(Image.open(img_path))
  # train_dataset = image_dataset_from_directory(
  #     os.path.join(dataset_path, "train"),
  #     shuffle=True,
  #     batch_size=BATCH_SIZE,
  #     image_size=IMG_SIZE
  #   )

  # test_dataset = image_dataset_from_directory(
  #     os.path.join(dataset_path, "test"),
  #     shuffle=True,
  #     batch_size=BATCH_SIZE,
  #     image_size=IMG_SIZE
  #   )

  #train_dataset = train_dataset.prefetch(4)
  # test_dataset = test_dataset.prefetch(4)

  # base_model_uri = "https://tfhub.dev/tensorflow/centernet/resnet50v2_512x512/1"
  base_model_uri = "https://tfhub.dev/tensorflow/efficientdet/lite4/feature-vector/2"

  base_model = hub.KerasLayer(base_model_uri, trainable=False)

  #base_model.trainable = False

  inputs = tf.keras.Input(shape=(640,640,3))
  x = base_model(inputs) #base_model(inputs, training=False)
  model= tf.keras.Model(inputs, x)

  # base_learning_rate = 0.0001
  # model.compile(optimizer= \
  #               tf.keras.optimizers.Adam(learning_rate=base_learning_rate),
  #               loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
  #               metrics=['accuracy'])

  print(model.summary())
















  # print(image.shape)

  # pred = base_model.predict(image[np.newaxis,...])

  # print(pred)

  #------------------------- 
  # Apply image detector on a single image.
  #detector = hub.load(base_model_uri)


  #global_average_layer = tf.keras.layers.GlobalAveragePooling2D()
  ## feature_batch_average = global_average_layer(feature_batch)

  #prediction_layer = tf.keras.layers.Dense(1)
  ## prediction_batch = prediction_layer(feature_batch_average)

  #inputs = tf.keras.Input(shape=(160,160,3))
  #x = data_augmentation(inputs)
  #x = preprocess_input(x)
  #x = base_model(x, training=False)
  #x = global_average_layer(x)
  #x = tf.keras.layers.Dropout(0.2)(x)
  #outputs = prediction_layer(x)
  #model = tf.keras.Model(inputs, outputs)

  #base_learning_rate = 0.0001
  #model.compile(optimizer= \
  #              tf.keras.optimizers.Adam(learning_rate=base_learning_rate),
  #              loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
  #              metrics=['accuracy'])















  #print(model.summary())














  # detector_output = detector(image[np.newaxis,...])

  # x = detector_output['detection_boxes'].numpy().astype(np.int32) #100 x 4
  # y = detector_output['detection_scores'].numpy()
  # z = detector_output['num_detections'].numpy().astype(np.int32)[0]

  #image_batch, label_batch = next(iter(train_dataset))
  ## feature_batch = base_model(image_batch)

  #base_model.trainable = False

  #global_average_layer = tf.keras.layers.GlobalAveragePooling2D()
  ## feature_batch_average = global_average_layer(feature_batch)

  #prediction_layer = tf.keras.layers.Dense(1)
  ## prediction_batch = prediction_layer(feature_batch_average)

  #inputs = tf.keras.Input(shape=(160,160,3))
  #x = data_augmentation(inputs)
  #x = preprocess_input(x)
  #x = base_model(x, training=False)
  #x = global_average_layer(x)
  #x = tf.keras.layers.Dropout(0.2)(x)
  #outputs = prediction_layer(x)
  #model = tf.keras.Model(inputs, outputs)

  #base_learning_rate = 0.0001
  #model.compile(optimizer= \
  #              tf.keras.optimizers.Adam(learning_rate=base_learning_rate),
  #              loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
  #              metrics=['accuracy'])

  #print(model.summary())

  #initial_epochs = 10
  #loss0, accuracy0 = model.evaluate(test_dataset)

  #print("initial loss: {:.2f}".format(loss0))
  #print("initial accuracy: {:.2f}".format(accuracy0))

  #history = model.fit(train_dataset,
  #                    epochs=initial_epochs,
  #                    validation_data=test_dataset)

  #model.save("no_fine_chkpt/")


  ###################################################################
  ## acc = history.history['accuracy']
  ## val_acc = history.history['val_accuracy']
  
  ## loss = history.history['loss']
  ## val_loss = history.history['val_loss']
  
  ## plt.figure(figsize=(8, 8))
  ## plt.subplot(2, 1, 1)
  ## plt.plot(acc, label='Training Accuracy')
  ## plt.plot(val_acc, label='Validation Accuracy')
  ## plt.legend(loc='lower right')
  ## plt.ylabel('Accuracy')
  ## plt.ylim([min(plt.ylim()),1])
  ## plt.title('Training and Validation Accuracy')
  
  ## plt.subplot(2, 1, 2)
  ## plt.plot(loss, label='Training Loss')
  ## plt.plot(val_loss, label='Validation Loss')
  ## plt.legend(loc='upper right')
  ## plt.ylabel('Cross Entropy')
  ## plt.ylim([0,1.0])
  ## plt.title('Training and Validation Loss')
  ## plt.xlabel('epoch')
  ## plt.show()

  ###################################################################

  #fine_tune_at = 100

  #base_model.trainable = True

  #for layer in base_model.layers[:fine_tune_at]:
  #  layer.trainable = False


  #model.compile(loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
  #              optimizer = tf.keras.optimizers.RMSprop(
  #                  learning_rate=base_learning_rate / 10),
  #              metrics=['accuracy'])

  #print(model.summary())

  #fine_tune_epochs = 10
  #total_epochs = initial_epochs + fine_tune_epochs

  #history_fine = model.fit(train_dataset,
  #                         epochs=total_epochs,
  #                         initial_epoch=history.epoch[-1],
  #                         validation_data=test_dataset)

  #model.save("fine_chkpt/")

  #loss, accuracy = model.evaluate(test_dataset)
  #print('loss = {},\naccuracy = {}'.format(loss, accuracy))




import os
import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import ganondorf.data
from ganondorf.pix2pix import residual
from ganondorf.pix2pix import blocks

from tensorflow_examples.models.pix2pix import pix2pix
from tensorflow.keras.preprocessing import image_dataset_from_directory

simage = None
smask = None

# def nearest_convolve(arr):
#   out = np.empty_like(arr)
#   height = arr.shape[0]
#   width = arr.shape[1]
#   for i in range(height):
#     for j in range(width):
      


def add_sample_weights(image, mask):
  class_weights = tf.constant([1.33, 1.0]) # 1.33 looks awesome i think 1.25 too
  class_weights = class_weights / tf.reduce_sum(class_weights)

  sample_weights = tf.gather(class_weights, indices=tf.cast(mask, tf.int32))

  return image, mask, sample_weights

def normalize(tensor_image):
  return tf.cast(tensor_image, tf.float32) / 255.0

def display(display_list):
  #plt.figure(figsize=(15,15))
  title = ['Input Image', 'True Mask', 'Predicted Mask']

  for i in range(len(display_list)):
    plt.subplot(1, len(display_list), i+1)
    plt.title(title[i])
    plt.imshow(tf.keras.preprocessing.image.array_to_img(display_list[i]))
    plt.axis('off')
  plt.show()

@tf.function
def load_image_train(image, mask, image_shape=(128,128)):
  #input_image = tf.image.resize(image, image_shape)
  #input_mask  = tf.image.resize(mask, image_shape)

  if tf.random.uniform(()) > 0.5:
    # input_image = tf.image.flip_left_right(image)#input_image)
    # input_mask  = tf.image.flip_left_right(mask)#input_mask)
    image = tf.image.flip_left_right(image)#input_image)
    mask  = tf.image.flip_left_right(mask)#input_mask)

  # input_image = normalize(input_image)
  image = normalize(image)

  # return input_image, input_mask
  return image, mask

# Why no @tf.function?
def load_image_test(image, mask, image_shape=(128,128)):
  #input_image = tf.image.resize(image, (128,128))
  #input_mask  = tf.image.resize(mask, image_shape)

  input_image = normalize(image)#input_image)

  return input_image, mask #input_mask
  

def unet_model(output_channels, down_stack, up_stack, res_stack=None):
  inputs = tf.keras.layers.Input(shape=[128,128,3])

  skips = down_stack(inputs)
  x = skips[-1]
  skips = reversed(skips[:-1])

  if res_stack is not None:
    for res in res_stack:
      x = res(x)

  for up, skip in zip(up_stack, skips):
    x = up(x)
    concat = tf.keras.layers.Concatenate()
    x = concat([x, skip])

  last = tf.keras.layers.Conv2DTranspose(output_channels,
                                         3,
                                         strides=2,
                                         padding='same')
  x = last(x)

  return tf.keras.Model(inputs=inputs, outputs=x)

def create_mask(pred_mask):
  pred_mask = tf.argmax(pred_mask, axis=-1) # Choose arg with max probability
  pred_mask = pred_mask[..., tf.newaxis]
  return pred_mask[0]

def show_predictions(model, dataset=None, num=1, simage=None, smask=None):
  if dataset:
    for image, mask in dataset.take(num):
      pred_mask = model.predict(image)
      display([image[0], mask[0], create_mask(pred_mask)])
  else:
    display([simage, 
             smask, 
             create_mask(model.predict(simage[tf.newaxis, ...]))])



class DisplayCallback(tf.keras.callbacks.Callback):
  def on_epoch_end(self, epoch, logs=None):
    #clear_output(wait=True)
    if (epoch + 1) % 10 == 0:
      show_predictions(self.model, simage=simage, smask=smask)
    print('\nSample Predictions after epoch {}\n'.format(epoch+1))


if __name__ == '__main__':

  # DATASET_PATH = os.path.join(
  #     "..","data","datasets","datasets","ALSegment"
  #     )

  # TRAIN_PATH = os.path.join(DATASET_PATH, "train")
  # TEST_PATH  = os.path.join(DATASET_PATH, "test")

  TRAIN_LENGTH = 34 #len(os.listdir(TRAIN_PATH))
  BATCH_SIZE = 64
  BUFFER_SIZE = 1000
  STEPS_PER_EPOCH = 1#TRAIN_LENGTH // BATCH_SIZE #  will be 0, is this an issue?

  # train_data = segloader.load_segmentation(TRAIN_PATH)

  # test_data = segloader.load_segmentation(TEST_PATH)

  train_data, test_data = ganondorf.data.Dataset.load("ALSegmentation",
                                                      size=(128,128))

  train = train_data.map(load_image_train,
                            num_parallel_calls=tf.data.experimental.AUTOTUNE)

  # train = train_data.map(add_sample_weights,
                            # num_parallel_calls=tf.data.experimental.AUTOTUNE)

  test = test_data.map(load_image_test)

  train_dataset = train.cache().shuffle(BUFFER_SIZE).batch(BATCH_SIZE)#.repeat()
  train_dataset = train_dataset.prefetch(
      buffer_size=tf.data.experimental.AUTOTUNE
      )


  test_dataset = test.batch(BATCH_SIZE)

  # Test 1 works too nicly so uses test 2
  it = iter(test.take(2))
  next(it)
  simage, smask  = next(it)

  OUTPUT_CHANNELS = 2

  base_model = tf.keras.applications.MobileNetV2(input_shape=[128, 128, 3],
                                                 include_top=False)

  layer_names = [
      'block_1_expand_relu',  #  64x64x96
      'block_3_expand_relu',  #  32x32x144
      'block_6_expand_relu',  #  16x16x192
      'block_13_expand_relu', #  8x8x576
      'block_16_project',     #  4x4x320
      ]

  base_model_outputs = \
      [base_model.get_layer(name).output for name in layer_names]

  down_stack = tf.keras.Model(inputs=base_model.input,
                              outputs=base_model_outputs)

  down_stack.trainable = False

  residual_stack = [
      residual.ResidualBottleneckLayer.as_residual_bridge(2, 320),
      residual.ResidualBottleneckLayer.as_residual_bridge(2, 320),
      residual.ResidualBottleneckLayer.as_residual_bridge(2, 320),
      tf.keras.layers.ReLU(),
      ]

  # after_residual_activation = tf.keras.layers.ReLU()


  #Old Version
  # up_stack = [
  #     pix2pix.upsample(512, 3),
  #     pix2pix.upsample(256, 3),
  #     pix2pix.upsample(128, 3),
  #     pix2pix.upsample(64, 3),
  #     ]

  up_stack = [
      blocks.decoder_block_2D(512, name="decode_block_1"),
      blocks.decoder_block_2D(256, name="decode_block_2"),
      blocks.decoder_block_2D(128, name="decode_block_3"),
      blocks.decoder_block_2D(64,  name="decode_block_4"),
      ]


  model = unet_model(OUTPUT_CHANNELS,
                     down_stack, up_stack, res_stack=residual_stack)

  model.compile(optimizer='adam',
                loss=tf.keras.losses.SparseCategoricalCrossentropy(
                    from_logits=True
                    ),
                metrics=['accuracy'])

  tf.keras.utils.plot_model(model, show_shapes=True, dpi=64, to_file="plot.svg")

  show_predictions(model, simage=simage, smask=smask)


  EPOCHS = 78
  VAL_SUBSPLITS = 5
  VALIDATION_STEPS = 1#len(os.listdir(TEST_PATH)) // BATCH_SIZE // VAL_SUBSPLITS

  model_history = model.fit(train_dataset.map(add_sample_weights),
                            epochs=EPOCHS,
                            #steps_per_epoch=STEPS_PER_EPOCH,
                            #validation_steps=VALIDATION_STEPS,
                            validation_data=test_dataset,
                            callbacks=[DisplayCallback()])

  loss = model_history.history['loss']
  val_loss = model_history.history['val_loss']

  plt.figure()
  plt.plot(model_history.epoch, loss, 'r', label='Training loss')
  plt.plot(model_history.epoch, val_loss, 'bo', label='Validation loss')
  plt.title('Training and Validation Loss')
  plt.xlabel('Epoch')
  plt.ylabel('Loss Value')
  plt.ylim([0, 1])
  plt.legend()
  plt.show()

  # model.save("chkpt/")
  model.save("E{:02}_chkpt/".format(EPOCHS))





import sys

import PIL
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt


if __name__ == '__main__':
  if len(sys.argv) < 3:
    print("Requires path to image being passed in."
          " and path to model too."
          " Exiting ....")

  img = PIL.Image.open(sys.argv[1]).convert('RGB').resize((128,128))
  arr = np.asarray(img)[np.newaxis,...]

  model = tf.keras.models.load_model(sys.argv[2])
  predict = model.predict(arr)[0]
  predict = np.argmax(predict, axis=-1)

  plt.imshow(predict)
  plt.show()


  




""" DOC STRING """
import os
import math
import csv
import random
from multiprocessing import Pool
from functools import partial
from typing import Tuple, Dict, Sequence, Iterable, List
import numpy as np

BASE_DATASET_DIR = os.path.dirname(os.path.realpath(__file__)) + "/dataset"

strip_extension = lambda a: a[:a.rindex(".")]

def csv_line_to_typed_array(csv_line, ptype: type = float,
                            dtype: type = np.float32):
  array_fn = partial(np.array, dtype=dtype)
  return array_fn(list(map(ptype, csv_line)))

def csv_to_array_list(filenames: Iterable[str],
                       fieldnames: Sequence[str] = None,
                       dataset_directory: str = BASE_DATASET_DIR,
                       ptype: type = float, dtype: type = np.float32,
                       num_processes: int = None) -> List[np.array]:
  #array_list = []
  csv_map_fn = partial(csv_to_array,
                       fieldnames=fieldnames,
                       dataset_directory=dataset_directory, ptype=ptype,
                       dtype=dtype, num_processes=num_processes)

  return list(map(csv_map_fn, filenames))


def csv_to_array(filename: str, fieldnames: Sequence[str] = None,
                 dataset_directory: str = BASE_DATASET_DIR,
                 ptype: type = float, dtype: type = np.float32,
                 num_processes: int = None) -> np.array:
  arrays = None
  csv_map_fn = partial(csv_line_to_typed_array, ptype=ptype, dtype=dtype)
  with open("{}/{}".format(dataset_directory, filename)) as csv_file:
    csv_reader = csv.DictReader(csv_file, fieldnames=fieldnames)
    lines = (list(line.values()) for line in csv_reader)
    with Pool(num_processes) as p:
      arrays = p.map(csv_map_fn, lines)

    # for line in csv_reader:
      # arrays.append(np.array(list(map(float, line.values())), dtype=dtype))

  return np.vstack(arrays)

def split_client_data(input_data: Sequence[np.array], classes: Sequence,
                      clientnames: Sequence[str], label_index: int = -1,
                      test_ratio: float = 0.1, randomize: bool = True,
                      num_processes: int = None) -> Dict[str, np.array]:
  dataset = {}
  split_map_fn = partial(split_test_and_train, classes=classes,
                         label_index=label_index, test_ratio=test_ratio,
                         randomize=randomize)
  client_data = None
  with Pool(num_processes) as p:
    client_data = p.map(split_map_fn, input_data)

  for client_index, clientname in enumerate(clientnames):
    dataset[clientname + "_train_data"] = client_data[client_index][0]
    dataset[clientname + "_test_data"] = client_data[client_index][1]

  return dataset


def split_test_and_train(client_data: np.array, classes: Sequence,
                         label_index: int = -1, test_ratio: float = 0.1,
                         randomize: bool = True) -> Tuple[np.array, np.array]:
  train_data_assigned = False
  test_data_assigned = False

  train_data = None
  test_data = None

  for clazz in classes:
    #get all the data for a client with label == clazz
    data = np.array([data for data in client_data
                     if data[label_index] == clazz])

    instance_count = data.shape[0]
    test_indicies_count = math.ceil(instance_count * test_ratio)
    test_indicies = None

    if randomize:
      test_indicies = random.choices(range(instance_count),
                                     k = test_indicies_count)
    else:
      test_indicies = range(instance_count)[-test_indicies_count:]

    for i, arr in enumerate(data):
      if i in test_indicies:
        if not test_data_assigned:
          test_data = arr
          test_data_assigned = True
        else:
          test_data = np.vstack((test_data, arr))
      else:
        if not train_data_assigned:
          train_data = arr
          train_data_assigned = True
        else:
          train_data = np.vstack((train_data, arr))

  return (train_data, test_data)


def combine_npz_arrays(filenames: Iterable[str]) -> Sequence[np.array]:
  data_for_clients = []
  for fname in filenames:
    npz = np.load(fname)
    client_data = np.hstack([npz[f] for f in npz.files])
    data_for_clients.append(client_data)

  return data_for_clients



